/*
 * Copyright 2008-2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.derquinsej;

import static com.google.common.base.Preconditions.checkNotNull;
import static java.lang.System.nanoTime;
import static net.sf.derquinsej.stats.Timings.createAtomic;

import java.util.Locale;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.TimeUnit;

import net.sf.derquinsej.stats.AtomicTiming;
import net.sf.derquinsej.stats.Timing;

import com.google.common.base.Function;
import com.google.common.collect.MapMaker;

/**
 * A simple caching function backed by a computing concurrent map.
 * @author Andres Rodriguez
 * @param <F> The type of the function input.
 * @param <T> The type of the function output.
 */
public final class MapCachingFunction<F, T> implements CachingFunction<F, T> {
	/**
	 * Returns a new builder.
	 * @return A new builder.
	 */
	public static final Builder builder() {
		return new Builder();
	}

	/** Backing map. */
	private final ConcurrentMap<F, T> map;
	/** Computer. */
	private final Computer computer;
	/** Request timing. */
	private final AtomicTiming request = createAtomic(TimeUnit.MILLISECONDS);
	/** Computation timing. */
	private final AtomicTiming computation = createAtomic(TimeUnit.MILLISECONDS);

	/**
	 * Constructor.
	 * @param maker Map maker.
	 * @param function Function to apply.
	 */
	private MapCachingFunction(MapMaker maker, Function<? super F, ? extends T> function) {
		this.computer = new Computer(function);
		this.map = maker.makeComputingMap(computer);
	}

	/*
	 * (non-Javadoc)
	 * @see com.google.common.base.Function#apply(java.lang.Object)
	 */
	public T apply(F from) {
		final long t0 = nanoTime();
		final T value = map.get(from);
		request.add((nanoTime() - t0) / 1000000);
		return value;
	};

	/*
	 * (non-Javadoc)
	 * @see net.sf.derquinsej.CachingFunction#getRequested()
	 */
	public long getRequested() {
		return request.get().getCount();
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.derquinsej.CachingFunction#getComputed()
	 */
	public long getComputed() {
		return computation.get().getCount();
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.derquinsej.CachingFunction#getHitRatio()
	 */
	public double getHitRatio() {
		final long r = getRequested();
		return (r == 0) ? 0.0 : ((double) (r - getComputed()) * 100) / r;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.derquinsej.CachingFunction#getComputationTiming()
	 */
	public Timing getComputationTiming() {
		return computation.get();
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.derquinsej.CachingFunction#getRequestTiming()
	 */
	public Timing getRequestTiming() {
		return request.get();
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return String.format((Locale) null, "Function [%s]: %f %% HR, R:%s, C:%s", computer.function.toString(),
				getHitRatio(), getRequestTiming(), getComputationTiming());
	}

	/**
	 * Map entry computer function.
	 * @param <F> The type of the function input.
	 * @param <T> The type of the function output.
	 */
	private class Computer implements Function<F, T> {
		/** Function to apply. */
		private final Function<? super F, ? extends T> function;

		/**
		 * Constructor.
		 * @param function Function to apply.
		 */
		private Computer(Function<? super F, ? extends T> function) {
			this.function = function;
		}

		public T apply(F from) {
			final long t0 = nanoTime();
			final T value = function.apply(from);
			computation.add((nanoTime() - t0) / 1000000);
			return value;
		}

	}

	/**
	 * Builder for caching function backed by a computing concurrent map.
	 * @author Andres Rodriguez
	 * @param <F> The type of the function input.
	 * @param <T> The type of the function output.
	 */
	public static class Builder {
		/** Map maker. */
		private final MapMaker maker = new MapMaker();

		/** Constructor. */
		private Builder() {
		}

		/**
		 * Specifies that each key (not value) stored in the map should be
		 * wrapped in a SoftReference (by default, strong references are used).
		 * @return This builder.
		 * @throws IllegalStateException if the key strength was already set
		 */
		public Builder softKeys() {
			maker.softKeys();
			return this;
		}

		/**
		 * Specifies that each value (not key) stored in the map should be
		 * wrapped in a SoftReference (by default, strong references are used).
		 * @return This builder.
		 * @throws IllegalStateException if the value strength was already set
		 */
		public Builder softValues() {
			maker.softValues();
			return this;
		}

		/**
		 * Specifies that each entry should be automatically removed from the
		 * map once a fixed duration has passed since the entry's creation.
		 * @param duration The length of time after an entry is created that it
		 *            should be automatically removed.
		 * @param unit The unit that duration is expressed in.
		 * @return This builder.
		 * @throws IllegalArgumentException if the duration is not positive.
		 * @throws IllegalStateException if the expiration time was already set
		 */
		public Builder expiration(long duration, TimeUnit unit) {
			maker.expiration(duration, unit);
			return this;
		}

		/**
		 * Builds the caching function.
		 * @param function Function to cache.
		 * @return The caching function.
		 */
		public <F, T> MapCachingFunction<F, T> build(Function<? super F, ? extends T> function) {
			return new MapCachingFunction<F, T>(maker, checkNotNull(function));
		}

	}

}
