/*
 * Copyright 2008-2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.derquinsej.tuple;

/**
 * Helper functions to operate on tuples.
 * @author Andres Rodriguez
 */
public final class Tuples {
	/** Not instantiable. */
	private Tuples() {
		throw new AssertionError();
	}
	
	/**
	 * Extractor.
	 * @param t Tuple.
	 * @param i Index.
	 * @return The requested element or {@code null} if the argument is {@code null}.
	 */
	@SuppressWarnings("unchecked")
	private static <T>  T e(Tuple t, int i) {
		if (t == null) {
			return null;
		}
		return (T)t.get(i);
	}
	

	/**
	 * Returns the empty tuple.
	 * @return The empty tuple.
	 */
	public static Tuple tuple0() {
		return Tuple0Impl.TUPLE0;
	}

	/**
	 * Builds a 1-element tuple.
	 * @param e1 First element.
	 * @return The requested tuple.
	 */
	public static <T1> Tuple1<T1> tuple1(T1 e1) {
		return new Tuple1Impl<T1>(e1);
	}
	
	/**
	 * Builds a 2-element tuple.
	 * @param e1 First element.
	 * @param e2 Second element.
	 * @return The requested tuple.
	 */
	public static <T1, T2> Tuple2<T1, T2> tuple2(T1 e1, T2 e2) {
		return new Tuple2Impl<T1, T2>(e1, e2);
	}

	/**
	 * Builds a 2-element tuple.
	 * @param t1 First element.
	 * @param e2 Second element.
	 * @return The requested tuple.
	 */
	public static <T1, T2> Tuple2<T1, T2> tuple2(Tuple1<T1> t1, T2 e2) {
		return tuple2(Tuples.<T1>e(t1, 0), e2);
	}

	/**
	 * Builds a 2-element tuple.
	 * @param e1 First element.
	 * @param t2 Second element.
	 * @return The requested tuple.
	 */
	public static <T1, T2> Tuple2<T1, T2> tuple2(T1 e1, Tuple1<T2> t2) {
		return tuple2(e1, Tuples.<T2>e(t2, 0));
	}

	/**
	 * Builds a 2-element tuple.
	 * @param t1 First element.
	 * @param t2 Second element.
	 * @return The requested tuple.
	 */
	public static <T1, T2> Tuple2<T1, T2> tuple2(Tuple1<T1> t1, Tuple1<T2> t2) {
		return tuple2(Tuples.<T1>e(t1, 0), Tuples.<T2>e(t2, 0));
	}

	/**
	 * Builds a 3-element tuple.
	 * @param e1 First element.
	 * @param e2 Second element.
	 * @param e3 Third element.
	 * @return The requested tuple.
	 */
	public static <T1, T2, T3> Tuple3<T1, T2, T3> tuple3(T1 e1, T2 e2, T3 e3) {
		return new Tuple3Impl<T1, T2, T3>(e1, e2, e3);
	}

	/**
	 * Builds a 3-element tuple.
	 * @param t1 First element.
	 * @param e2 Second element.
	 * @param e3 Third element.
	 * @return The requested tuple.
	 */
	public static <T1, T2, T3> Tuple3<T1, T2, T3> tuple3(Tuple1<T1> t1, T2 e2, T3 e3) {
		return tuple3(Tuples.<T1>e(t1, 0), e2, e3);
	}

	/**
	 * Builds a 3-element tuple.
	 * @param e1 First element.
	 * @param t2 Second element.
	 * @param e3 Third element.
	 * @return The requested tuple.
	 */
	public static <T1, T2, T3> Tuple3<T1, T2, T3> tuple3(T1 e1, Tuple1<T2> t2, T3 e3) {
		return tuple3(e1, Tuples.<T2>e(t2, 0), e3);
	}

	/**
	 * Builds a 3-element tuple.
	 * @param e1 First element.
	 * @param e2 Second element.
	 * @param t3 Third element.
	 * @return The requested tuple.
	 */
	public static <T1, T2, T3> Tuple3<T1, T2, T3> tuple3(T1 e1, T2 e2, Tuple1<T3> t3) {
		return tuple3(e1, e2, Tuples.<T3>e(t3, 0));
	}
	
	/**
	 * Builds a 3-element tuple.
	 * @param t1 First element.
	 * @param t2 Second element.
	 * @param e3 Third element.
	 * @return The requested tuple.
	 */
	public static <T1, T2, T3> Tuple3<T1, T2, T3> tuple3(Tuple1<T1> t1, Tuple1<T2> t2, T3 e3) {
		return tuple3(Tuples.<T1>e(t1, 0), Tuples.<T2>e(t2, 0), e3);
	}

	/**
	 * Builds a 3-element tuple.
	 * @param t1 First element.
	 * @param e2 Second element.
	 * @param t3 Third element.
	 * @return The requested tuple.
	 */
	public static <T1, T2, T3> Tuple3<T1, T2, T3> tuple3(Tuple1<T1> t1, T2 e2, Tuple1<T3> t3) {
		return tuple3(Tuples.<T1>e(t1, 0), e2, Tuples.<T3>e(t3, 0));
	}

	/**
	 * Builds a 3-element tuple.
	 * @param e1 First element.
	 * @param t2 Second element.
	 * @param t3 Third element.
	 * @return The requested tuple.
	 */
	public static <T1, T2, T3> Tuple3<T1, T2, T3> tuple3(T1 e1, Tuple1<T2> t2, Tuple1<T3> t3) {
		return tuple3(e1, Tuples.<T2>e(t2, 0), Tuples.<T3>e(t3, 0));
	}

	/**
	 * Builds a 3-element tuple.
	 * @param t1 First element.
	 * @param t2 Second element.
	 * @param t3 Third element.
	 * @return The requested tuple.
	 */
	public static <T1, T2, T3> Tuple3<T1, T2, T3> tuple3(Tuple1<T1> t1, Tuple1<T2> t2, Tuple1<T3> t3) {
		return tuple3(Tuples.<T1>e(t1, 0), Tuples.<T2>e(t2, 0), Tuples.<T3>e(t3, 0));
	}
	
	/**
	 * Builds a 3-element tuple.
	 * @param t1_2 Elements 1-2.
	 * @param e3 Third element.
	 * @return The requested tuple.
	 */
	public static <T1, T2, T3> Tuple3<T1, T2, T3> tuple3(Tuple2<T1, T2> t1_2, T3 e3) {
		return tuple3(Tuples.<T1>e(t1_2, 0), Tuples.<T2>e(t1_2, 1), e3);
	}

	/**
	 * Builds a 3-element tuple.
	 * @param t1_2 Elements 1-2.
	 * @param t3 Third element.
	 * @return The requested tuple.
	 */
	public static <T1, T2, T3> Tuple3<T1, T2, T3> tuple3(Tuple2<T1, T2> t1_2, Tuple1<T3> t3) {
		return tuple3(t1_2, Tuples.<T3>e(t3, 0));
	}

	/**
	 * Builds a 3-element tuple.
	 * @param e1 First element.
	 * @param t2_3 Elements 2-3.
	 * @return The requested tuple.
	 */
	public static <T1, T2, T3> Tuple3<T1, T2, T3> tuple3(T1 e1, Tuple2<T2, T3> t2_3) {
		return tuple3(e1, Tuples.<T2>e(t2_3, 0), Tuples.<T3>e(t2_3, 1));
	}
	
	/**
	 * Builds a 3-element tuple.
	 * @param t1 First element.
	 * @param t2_3 Elements 2-3.
	 * @return The requested tuple.
	 */
	public static <T1, T2, T3> Tuple3<T1, T2, T3> tuple3(Tuple1<T1> t1, Tuple2<T2, T3> t2_3) {
		return tuple3(Tuples.<T1>e(t1, 0), t2_3);
	}
	
}
